use clap::Parser;
use colored::Colorize;
use nix::unistd::Uid;
use nvml_wrapper::Nvml;
use std::process::exit;

#[derive(Parser, Debug)]
struct TargetValues {
    /// The power limit the GPU should target (in Watts)
    #[arg(short, long)]
    power: Option<i32>,
    /// The clock offset the GPU should apply (in Hz)
    #[arg(short, long)]
    clock_offset: Option<i32>,
    /// The memory clock offset the GPU should apply (in Hz)
    #[arg(short, long)]
    mem_clock_offset: Option<i32>,
}

impl TargetValues {
    fn all_match_condition<F>(&self, condition: F) -> bool
    where
        F: Fn(Option<i32>) -> bool,
    {
        condition(self.power) && condition(self.mem_clock_offset) && condition(self.clock_offset)
    }

    fn all_empty(&self) -> bool {
        self.all_match_condition(|x: Option<i32>| x.is_none())
    }
}

fn to_milliwatts(watts: i32) -> i32 {
    watts * 1000
}

fn is_ran_as_root() -> bool {
    Uid::effective().is_root()
}

fn main() {
    let args = TargetValues::parse();

    match is_ran_as_root() {
        true => println!("{}", "Running as root!".green()),
        false => {
            println!("{}", "This script requires to be ran as root!".red());
            exit(1);
        }
    }

    let nvml = Nvml::init().unwrap();
    let mut device = nvml.device_by_index(0).unwrap();

    println!("GPU name: {:?}", device.name().unwrap());

    if args.all_empty() {
        println!("There is nothing to do!");
        exit(1);
    }

    if let Some(power) = args.power {
        println!(
            "{}",
            format!("Trying to target {}W power limit.", power).cyan()
        );
        match device.set_power_management_limit(to_milliwatts(power).try_into().unwrap()) {
            Ok(_) => println!("{}", format!("Target {}W power limit.", power).green()),
            _ => println!("{}", "An error occured.".red()),
        }
    }
    println!(
        "Effective target power limit: {:?}",
        device.power_management_limit().unwrap()
    );

    if let Some(clock_offset) = args.clock_offset {
        println!(
            "{}",
            format!("Trying to target {}Hz GPU clock offset.", clock_offset).cyan()
        );
        match device.set_gpc_clock_vf_offset(clock_offset) {
            Ok(_) => println!(
                "{}",
                format!("Target {}Hz GPU clock offset.", clock_offset).green()
            ),
            _ => println!("{}", "An error occured.".red()),
        }
    }
    println!(
        "Effective target clock offset: {:?}",
        device.gpc_clock_vf_offset().unwrap()
    );

    if let Some(mem_clock_offset) = args.mem_clock_offset {
        println!(
            "{}",
            format!(
                "Trying to target {}Hz GPU memclock offset.",
                mem_clock_offset
            )
            .cyan()
        );
        match device.set_mem_clock_vf_offset(mem_clock_offset) {
            Ok(_) => println!(
                "{}",
                format!("Target {}Hz GPU memclock offset.", mem_clock_offset).green()
            ),
            _ => println!("{}", "An error occured.".red()),
        }
    }
    println!(
        "Effective target memory clock offset: {:?}",
        device.mem_clock_vf_offset().unwrap()
    );
}
